"Common functions used by show*.py utilities"

__author__ = "Paresh Adhia"
__copyright__ = "Copyright 2016-2017, Paresh Adhia"
__license__ = "GPL"

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import namedtuple
import tdtypes as td
from .. import util

logger = None

def args_size(p):
	"add size args"
	x = p.add_mutually_exclusive_group()
	x.add_argument("-h", "--human-readable", dest='sizefmt', action="store_const", const='.1h', default=',d',
		help="print human readable sizes (e.g., 1K 234M 2G)")
	x.add_argument("--si", dest='sizefmt', action="store_const", const='.1s', help="likewise, but use powers of 1000 not 1024")

def args_misc(p):
	"add misc args"
	g = p.add_argument_group("Miscellaneous")
	g.add_argument('-v', '-l', '--verbose', default=0, action='count', help='show verbose output. Use -vv for more details')
	g.add_argument('--help', action='help', help='show this help message and exit')
	td.dbconn_args(g)

def args_sort(p):
	"add ordering results args"
	x = p.add_mutually_exclusive_group()
	x.add_argument(      '--sort', metavar='WORD', default='name', choices=['none','size','name','time'],
		help="sort by WORD instead of name: none (-U), size (-S), time (-t)")
	x.add_argument("-t", dest='sort', action='store_const', const='mtime', help="sort by time, newest first")
	x.add_argument("-U", dest='sort', action='store_const', const='none',  help="do not sort; list entries in database order")
	x.add_argument("-S", dest='sort', action='store_const', const='size',  help="sort by file size, largest first")
	p.add_argument("-r", "--reverse", action="store_true",                 help="reverse order while sorting")

Col = namedtuple('Col', ['name', 'hdr', 'verb'])

class Listing:
	"Information listing"

	def __init__(self, name, doc):
		import argparse
		global logger

		logger = util.getLogger(name)

		self.p = argparse.ArgumentParser(description=doc, fromfile_prefix_chars='@', add_help=False)

	def ls(self):
		"script entry-point"
		for fn in self.user_args():
			fn(self.p)
		args_misc(self.p)

		args = self.p.parse_args()

		Col.sizefmt = args.sizefmt
		cols = [c for c in self.all_cols(args) if c.verb <= args.verbose]

		with td.cursor(args) as csr:
			sql = self.build_sql(args.names, ('{c.name} AS "{c.hdr}"'.format(c=c) for c in cols), args)
			logger.debug(sql)
			csr.execute(sql)
			if len(cols) > 1:
				util.pprint_csr(csr, sizefmt=getattr(args, 'sizefmt', '.1h'))
			else:
				for c, in csr.fetchall():
					print(c)

	def user_args(self):
		"returns a list of functions that will add arguments in to passed argparse Parser object"
		return []

	def build_sql(self, names, cols, opts):
		"build SQL based on run-time options"
		pass

	def all_cols(self, opts):
		"all possible column information that can be displayed"
		pass
